#!/usr/bin/env ruby

module IOExt
	def readb; read(1).unpack("C")[0]; end
	def readw; read(2).unpack("v")[0]; end
	def readd; read(4).unpack("V")[0]; end
end

class Struc
	def initialize arr
		@fmt = arr
		@klass = Struct.new(*@fmt.map{|e| e[1] }.select{|e| e })
	end
	def read io
		io.extend IOExt
		@klass.new *@fmt.map { |e|
			# We must read it to skip bytes (=side effect is required!)
			v = case e[0]
			when :u8; io.readb
			when :u16; io.readw
			when :u32; io.readd
			else; raise "?unknown struc field type #{e[0]}:#{e[1]}"
			end
			# Then use or throw it.
			e[1] ? v : nil
		}.select{|e| e }
	end
end

BMPFileHeader = Struc.new [
	[:u16, :type],
	[:u32, :size],
	[:u16, nil],
	[:u16, nil],
	[:u32, :offbits],
]
BMPInfoHeader = Struc.new [
	[:u32, :size],
	[:u32, :width],
	[:u32, :height],
	[:u16, :planes],
	[:u16, :bitcount],
	[:u32, :compression],
	[:u32, :sizeimage],
	[:u32, :xppm],
	[:u32, :yppm],
	[:u32, :clrused],
	[:u32, :clrimportant],
]
BGRA = Struc.new [
	[:u8, :b],
	[:u8, :g],
	[:u8, :r],
	[:u8, nil],
]

class BMP
	attr_reader :width, :height, :bpp, :lines

	def initialize io
		@fileheader = BMPFileHeader.read io
		@fileheader.type == 0x4D42 or raise "?bmp signature"

		@infoheader = BMPInfoHeader.read io
		@infoheader.size == 40 or raise "?bmp unknown infoheader size"
		@infoheader.compression == 0 or raise "?bmp unknown compression"
		@infoheader.bitcount == 8 or raise "?bmp unsupported bitdepth"
		@infoheader.planes == 1 or raise "?bmp unsupported planes"

		@palette = 8 < @infoheader.bitcount ? nil : (2 ** @infoheader.bitcount).times.map {
			BGRA.read io
		}

		# for image interface
		@width = @infoheader.width
		@height = @infoheader.height
		@bpp = @infoheader.bitcount

		io.pos = @fileheader.offbits

		linebytes = @infoheader.width * @infoheader.bitcount / 8
		paddingbytes = (4 - linebytes % 4) % 4
		@lines = @infoheader.height.times.map {
			l = case @infoheader.bitcount
			when 8; io.read(linebytes).unpack("C*")
			else; raise "!bmp unsupported bitdepth"
			end
			io.pos += paddingbytes
			l
		}.reverse
	end
end

class Image2C
	attr_accessor :images

	class Image < Struct.new :x, :y, :width, :height, :number, :lines
		def initialize x, y, number, lines
			self.x = x
			self.y = y
			self.width = lines[0].size
			self.height = lines.size
			self.number = number
			self.lines = lines

			# validate
			lines.each { |ln| ln.size == self.width or raise "!line width is not unique: #{x},#{y}+#{width}x#{height}:#{number}" }
		end
	end
	State = Struct.new :startx, :starty, :width, :height

	Font = [
		[[true]*3, *([[true, false, true]]*3), [true]*3],
		[*([[false, false, true]]*5)],
		[[true]*3, [false, false, true], [true]*3, [true, false, false], [true]*3],
		[*([[true]*3, [false, false, true]]*2), [true]*3],
		[*([[true, false, true]]*2), [true]*3, *([[false, false, true]]*2)],
		[[true]*3, [true, false, false], [true]*3, [false, false, true], [true]*3],
		[[true]*3, [true, false, false], [true]*3, [true, false, true], [true]*3],
		[[true]*3, *([[false, false, true]]*4)],
		[*([[true]*3, [true, false, true]]*2), [true]*3],
		[[true]*3, [true, false, true], [true]*3, [false, false, true], [true]*3],
	]

	def initialize img
		@img = img
		@ctransp = img.lines[0][0]
		@cinvert = img.lines[0][1]
		@cblack = img.lines[0][2]
		@cwhite = img.lines[0][3]
		@cmarker = img.lines[0][4]

		analyze
	end
	def readnum x, y
		ret = nil
		loop {
			raw = 5.times.map { |yy| @img.lines[y + yy][x, 3].map{ |e| e == @cblack } }
			num = Font.each_with_index.detect(proc{[nil, nil]}) { |e| e[0] == raw }[1]
			num == nil and break
			ret = (ret || 0) * 10 + num
			x += 4
		}
		ret
	end
	def analyze
		@images = []
		state = []

		# skip first line
		(1...@img.height).each { |y|
			# check is it continue, remove if it completed (or invalid if height==0)
			state.select! { |st|
				if @img.lines[y][st.startx] == @cmarker
					# continues
					st.height += 1
					true
				else
					# ends; add to images if it is valid
					if 1 < st.height
						number = st.starty < 6 ? nil : readnum(st.startx, st.starty - 6)
						lines = @img.lines[st.starty + 1, st.height - 1].map{ |ln| ln[st.startx + 1, st.width - 1] }

						# validate
						valid_colors = [@ctransp, @cinvert, @cblack, @cwhite]
						ix = nil
						invalid = lines.each_with_index.detect(proc{nil}) { |ln,y| ix = ln.each_with_index.detect(proc{nil}) { |e,x| not valid_colors.include? e } }
						if invalid
							puts "/* image at #{st.startx},#{st.starty} contains invalid pixels at #{st.startx+1+ix[1]},#{st.starty+1+invalid[1]}; ignored. */"
						else
							@images << Image.new(st.startx + 1, st.starty + 1, number, lines)
						end
					end
					false
				end
			}
			# add new image(-like) lines as state
			prevismark = @img.lines[y][0]
			@img.lines[y].slice_before { |e, s|
				ismark = e == @cmarker
				r = prevismark != ismark
				prevismark = ismark
				r
			}.inject({x: 0, r: []}) { |r,e|
				r[:r] << [r[:x], e]
				r[:x] += e.size
				r
			}[:r].each { |x, run|
				run[0] == @cmarker or next
				2 < run.size or next
				state << State.new(x, y, run.size, 1)
			}
		}
	end
	def to_c
		#"#include <inttypes.h>\n\n" +
		@images.map{ |image|
			imname = "image_#{image.number || "x#{image.x}y#{image.y}"}"

			"#define #{imname}_width #{image.width}\n" +
			"#define #{imname}_height #{image.height}\n" +
			"const uint32_t #{imname}_rmask[] = {\n" +
			image.lines.each_slice(4).map { |lns|
				"\t" + lns.map { |ln|
					"0x%08X," % ln.reverse.inject(0) { |r,e|
						(r << 1) | ((e == @cblack || e == @cwhite) ? 1 : 0)
					}
				}.join(" ") + "\n"
			}.join +
			"};\n" +
			"const uint32_t #{imname}_bit[] = {\n" +
			image.lines.each_slice(4).map { |lns|
				"\t\t" + lns.map { |ln|
					"0x%08X," % ln.reverse.inject(0) { |r,e|
						(r << 1) | ((e == @cinvert || e == @cblack) ? 1 : 0)
					}
				}.join(" ") + "\n"
			}.join +
			"};\n" +
			""
		}.join
	end
end

if $0 == __FILE__
	img = open(ARGV[0], "rb") { |f| BMP.new f }
	puts Image2C.new(img).to_c
end
