#include <stdio.h>
#include <string.h>
#include <malloc.h>

// TODO support bad-sectors

size_t freadbytes(void *buf, size_t size, FILE *fp) {
	size_t read = 0;
	while(0 < size) {
		size_t r = fread(buf, 1, size, fp);
		if(r == 0) {
			break;
		}
		read += r;
		buf = (char*)buf + read;
		size -= r;
	}
	return read;
}

size_t fwritebytes(const void *buf, size_t size, FILE *fp) {
	size_t wrote = 0;
	while(0 < size) {
		size_t r = fwrite(buf, 1, size, fp);
		if(r == 0) {
			break;
		}
		wrote += r;
		buf = (const char*)buf + wrote;
		size -= r;
	}
	return wrote;
}

int main(int argc, char **argv) {
	char buf[0x80];
	char *filebuf;
	size_t srcfilesize;
	int numrequiredblocks;
	int numfreeblocks;
	int freeblocks[15];
	FILE *fp;

	if(argc < 4) {
		fputs("Insufficient arguments.", stderr);
		fprintf(stderr, "usage: %s <mcfile> <srcfile> <internalfilename>\n", argv[0]);
		return 1;
	}

	// read source file
	if((filebuf = malloc(120 * 1024)) == NULL) { // just allocate. it is freed by OS.
		perror("could not allocate buffer for source file");
		return 1;
	}
	if((fp = fopen(argv[2], "rb")) == NULL) {
		perror("could not open source file");
		return 1;
	}
	srcfilesize = freadbytes(filebuf, 120 * 1024, fp);
	if(ferror(fp)) {
		perror("could not read source file");
		return 1;
	}
	if(fclose(fp)) {
		perror("could not close source file");
		return 1;
	}

	numrequiredblocks = (srcfilesize + 0x2000 - 1) / 0x2000; // div by block-size (0x2000, 8KiB), round-up.

	// !QuickHack: patch block size in SC header here!
	filebuf[3] = numrequiredblocks;

	// stuff for block align
	for(int i = srcfilesize; i < numrequiredblocks * 0x2000; i++) {
		filebuf[i] = 0;
	}

	// open mc file
	if((fp = fopen(argv[1], "r+b")) == NULL) {
		perror("could not open mc file");
		return 1;
	}

	// verify source
	if(fseek(fp, 0, SEEK_SET) < 0) {
		perror("could not seek mc file (to get magic)");
		return 1;
	}
	if(freadbytes(buf, 2, fp) != 2) {
		perror("could not read mc magic");
		return 1;
	}
	if(memcmp(buf, "MC", 2) != 0) {
		perror("mc magic is invalid: you are passing wrong file as mc file?");
		return 1;
	}

	// check is there enough room
	if(fseek(fp, 0x80, SEEK_SET) < 0) {
		perror("could not seek mc file (to check free space)");
		return 1;
	}
	for(int i = 1; i <= 15; i++) {
		if(freadbytes(buf, 0x80, fp) != 0x80) {
			perror("could not read mc directory entry");
			fprintf(stderr, "  for block %d.\n", i);
			return 1;
		}
		if((buf[0] & 0xF0) == 0x50) {
			/* using blocks */
		} else if((buf[0] & 0xF0) == 0xA0) {
			freeblocks[numfreeblocks] = i;
			numfreeblocks++;
		} else {
			fprintf(stderr, "error: bad status (%02X) at directory entry of block %d\n", buf[0], i);
			return 1;
		}
	}
	if(numfreeblocks < numrequiredblocks) {
		fprintf(stderr, "error: %d blocks free is not enough for storing %d blocks.\n", numfreeblocks, numrequiredblocks);
		return 1;
	}

	// write blocks
	for(int i = 0; i < numrequiredblocks; i++) {
		// directory entry
		bzero(buf, 0x80);
		if(i == 0) {
			buf[0] = 0x51; // first using block

			// using file size (LE)
			buf[4] = 0x00;
			buf[5] = (0x20 * numrequiredblocks) & 0xFF;
			buf[6] = ((0x20 * numrequiredblocks) >> 8) & 0xFF;
			buf[7] = 0x00;

			// internal file name: 20 bytes max (without null)
			strncpy(&buf[0x0A], argv[3], 20);
		} else if(i < numrequiredblocks - 1) {
			buf[0] = 0x52; // middle of using block
			// no filesize for middle block
			// no filename for middle block
		} else { // i == numrequiredblocks - 1
			buf[0] = 0x53; // last using block
			// no filesize for last block
			// no filename for last block
		}
		if(i < numrequiredblocks - 1) {
			// first or middle of blocks
			// note: next-block is 0-origin.
			buf[8] = freeblocks[i + 1] - 1;
			buf[9] = 0x00;
		} else {
			// last block
			// 0xFFFF next block = none
			buf[8] = 0xFF;
			buf[9] = 0xFF;
		}
		// checksum
		for(int i = 0; i < 0x7F; i++) {
			buf[0x7F] ^= buf[i];
		}

		if(fseek(fp, freeblocks[i] * 0x80, SEEK_SET) < 0) {
			perror("could not seek mc file for writing directory entry");
			fprintf(stderr, "  for block %d.\n", freeblocks[i]);
			return 1;
		}
		if(fwritebytes(buf, 0x80, fp) < 0) {
			perror("could not write directory entry to mc file");
			fprintf(stderr, "  for block %d.\n", freeblocks[i]);
			return 1;
		}

		// block itself
		if(fseek(fp, freeblocks[i] * 0x2000, SEEK_SET) < 0) {
			perror("could not seek mc file for writing a block");
			fprintf(stderr, "  at block %d.\n", freeblocks[i]);
			return 1;
		}
		if(fwritebytes(filebuf + i * 0x2000, 0x2000, fp) < 0) {
			perror("could not write a block to mc file");
			fprintf(stderr, "  at block %d.\n", freeblocks[i]);
			return 1;
		}
	}


	if(fclose(fp)) {
		perror("could not close mc file");
		return 1;
	}

	return 0;
}
